﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("GQAPNMoveEvaluator", "Evaluates an n-move.")]
  [StorableClass]
  public class GQAPNMoveEvaluator : SingleSuccessorOperator, IGQAPNMoveEvaluator {

    public ILookupParameter<IntegerVector> AssignmentParameter {
      get { return (ILookupParameter<IntegerVector>)Parameters["Assignment"]; }
    }
    public ILookupParameter<NMove> MoveParameter {
      get { return (ILookupParameter<NMove>)Parameters["Move"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ILookupParameter<DoubleValue> MoveInfeasibilityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveInfeasibility"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public ILookupParameter<DoubleValue> TransportationCostsParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public ILookupParameter<DoubleArray> DemandsParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ILookupParameter<DoubleArray> CapacitiesParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Capacities"]; }
    }

    [StorableConstructor]
    protected GQAPNMoveEvaluator(bool deserializing) : base(deserializing) { }
    protected GQAPNMoveEvaluator(GQAPNMoveEvaluator original, Cloner cloner) : base(original, cloner) { }
    public GQAPNMoveEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<IntegerVector>("Assignment", "The equipment-location assignment vector."));
      Parameters.Add(new LookupParameter<NMove>("Move", "The move to perform."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The solution quality."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The quality of the move if it would be applied."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveInfeasibility", "The infeasibility of the move if it would be applied."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights matrix describes the flows between the equipments."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances matrix describes the distances between the locations at which the equipment can be installed."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("InstallationCosts", "The installation costs matrix describes the installation costs of installing equipment i at location j"));
      Parameters.Add(new LookupParameter<DoubleValue>("TransportationCosts", "The transportation cost represents the flow-unit per distance-unit cost factor. This value can also be set to 1 if these costs are factored into the weights matrix already."));
      Parameters.Add(new LookupParameter<DoubleArray>("Demands", "The demands vector describes the space requirements of the equipments."));
      Parameters.Add(new LookupParameter<DoubleArray>("Capacities", "The capacities vector describes the available space at the locations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GQAPNMoveEvaluator(this, cloner);
    }

    public static DoubleValue Evaluate(NMove move, IntegerVector assignment, DoubleValue quality,
                                  DoubleArray demands, DoubleArray capacities,
                                  DoubleMatrix weights, DoubleMatrix distances,
                                  DoubleMatrix installationCosts,
                                  DoubleValue transportationCosts,
                                  out DoubleValue moveInfeasibility) {
      double moveQuality = quality.Value;
      int moves = move.N;
      var slack = (DoubleArray)capacities.Clone();
      Dictionary<int, int> moving = new Dictionary<int, int>();
      for (int i = 0; i < moves; i++) moving.Add(move.Equipments[i], move.Locations[i]);

      for (int i = 0; i < moves; i++) {
        int equip = move.Equipments[i];
        int newLoc = move.Locations[i];
        moveQuality -= installationCosts[equip, assignment[equip]];
        moveQuality += installationCosts[equip, newLoc];
        for (int j = 0; j < assignment.Length; j++) {
          if (!moving.ContainsKey(j)) {
            moveQuality += transportationCosts.Value * weights[equip, j] * (distances[newLoc, assignment[j]] - distances[assignment[equip], assignment[j]]);
            moveQuality += transportationCosts.Value * weights[j, equip] * (distances[assignment[j], newLoc] - distances[assignment[j], assignment[equip]]);
            if (i == 0) { // only once for each untouched equipment deduct the demand from the capacity
              slack[assignment[j]] -= demands[j];
            }
          } else {
            moveQuality += transportationCosts.Value * weights[equip, j] * (distances[newLoc, moving[j]] - distances[assignment[equip], assignment[j]]);
            moveQuality += transportationCosts.Value * weights[j, equip] * (distances[moving[j], newLoc] - distances[assignment[j], assignment[equip]]);
          }
        }
        slack[newLoc] -= demands[equip];
      }

      moveInfeasibility = new DoubleValue(-slack.Where(x => x < 0).Sum());
      return new DoubleValue(moveQuality);

    }

    public override IOperation Apply() {
      DoubleValue moveInfeasibility;
      MoveQualityParameter.ActualValue = Evaluate(MoveParameter.ActualValue,
                 AssignmentParameter.ActualValue, QualityParameter.ActualValue,
                 DemandsParameter.ActualValue, CapacitiesParameter.ActualValue,
                 WeightsParameter.ActualValue, DistancesParameter.ActualValue,
                 InstallationCostsParameter.ActualValue,
                 TransportationCostsParameter.ActualValue, out moveInfeasibility);
      MoveInfeasibilityParameter.ActualValue = moveInfeasibility;
      return base.Apply();
    }
  }
}
